/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

//
// Created by webpigeon on 18/04/22.
//

#include "fggl/gfx/ogl4/models.hpp"
#include "fggl/data/heightmap.h"

#include "fggl/gfx/camera.hpp"
#include <spdlog/spdlog.h>

namespace fggl::gfx::ogl4 {

	void StaticModelRenderer::resolveModels(ecs3::World &world) {
		// FIXME: this needs something reactive or performance will suck.
		auto renderables = world.findMatching<gfx::StaticMesh>();
		for (auto& renderable : renderables){
			auto currModel = world.get<StaticModel>( renderable );
			if ( currModel != nullptr ){
				continue;
			}

			auto* meshComp = world.get<gfx::StaticMesh>(renderable);

			// create a vao to store the mesh
			auto vao = std::make_shared< ogl::VertexArray >();
			vao->bind();

			// setup the mesh buffer stuff
			auto meshToUpload = meshComp->mesh;
			auto meshBuffer = std::make_shared<ogl::ArrayBuffer>();

			spdlog::info("mesh data to upload: {}", meshToUpload.vertexCount());
			meshBuffer->write(meshToUpload.vertexCount() * sizeof(data::Vertex), meshToUpload.vertexList().data(), ogl::BufUsage::STATIC_DRAW);

			// set up the vertex attributes
			auto posAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, posititon));
			auto normalAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, normal));
			auto colAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, colour));

			vao->setAttribute( *meshBuffer.get(), 0, posAttr );
			vao->setAttribute( *meshBuffer.get(), 1, normalAttr );
			vao->setAttribute( *meshBuffer.get(), 3, colAttr );

			// set up the element attributes
			auto elementBuffer = std::make_shared<ogl::ElementBuffer>();
			elementBuffer->write(meshToUpload.indexCount() * sizeof(uint32_t), meshToUpload.indexList().data(), ogl::BufUsage::STATIC_DRAW);

			auto* modelComp = world.add<StaticModel>(renderable);
			modelComp->vao = vao;
			modelComp->vertexData = meshBuffer;
			modelComp->elements = elementBuffer;
			modelComp->pipeline = m_phong;
			modelComp->elementCount = meshToUpload.indexCount();
			modelComp->drawType = ogl::Primative::TRIANGLE;

			// no active model, we need to resolve/load one.
			spdlog::info("looks like {} needs a static mesh", renderable);
		}

		// terrain
		auto terrain = world.findMatching<data::HeightMap>();
		for (auto& renderable : terrain){
			auto currModel = world.get<StaticModel>( renderable );
			if ( currModel != nullptr ){
				continue;
			}

			auto* heightmap = world.get<data::HeightMap>(renderable);

			// create a vao to store the mesh
			auto vao = std::make_shared< ogl::VertexArray >();
			vao->bind();

			// setup the mesh buffer stuff
			data::Mesh meshToUpload{};
			data::generateHeightMesh(heightmap, meshToUpload);
			auto meshBuffer = std::make_shared<ogl::ArrayBuffer>();

			spdlog::info("mesh data to upload: {}", meshToUpload.vertexCount());
			meshBuffer->write(meshToUpload.vertexCount() * sizeof(data::Vertex), meshToUpload.vertexList().data(), ogl::BufUsage::STATIC_DRAW);

			// set up the vertex attributes
			auto posAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, posititon));
			auto normalAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, normal));
			auto colAttr = ogl::attribute<data::Vertex, math::vec3>(offsetof(data::Vertex, colour));

			vao->setAttribute( *meshBuffer.get(), 0, posAttr );
			vao->setAttribute( *meshBuffer.get(), 1, normalAttr );
			vao->setAttribute( *meshBuffer.get(), 3, colAttr );

			// set up the element attributes
			auto elementBuffer = std::make_shared<ogl::ElementBuffer>();
			elementBuffer->write(meshToUpload.indexCount() * sizeof(uint32_t), meshToUpload.indexList().data(), ogl::BufUsage::STATIC_DRAW);

			auto* modelComp = world.add<StaticModel>(renderable);
			modelComp->vao = vao;
			modelComp->vertexData = meshBuffer;
			modelComp->elements = elementBuffer;
			modelComp->pipeline = m_phong;
			modelComp->elementCount = meshToUpload.indexCount();
			modelComp->drawType = ogl::Primative::TRIANGLE_STRIP;
			modelComp->restartIndex = meshToUpload.restartVertex;

			// no active model, we need to resolve/load one.
			spdlog::info("looks like heightmap, {} needs a static mesh", renderable);
		}
	}

	void StaticModelRenderer::renderModelsForward(const ecs3::World &world) {

		// fetch cameras we will need to render with
		auto cameras = world.findMatching<gfx::Camera>();

		// if there are no cameras, we can't do anything...
		if ( cameras.size() == 0) {
			spdlog::warn("asked to render static models, but there were no cameras");
			return;
		}

		float deltaTime = 0.0f;

		// perform a rendering pass for each camera (will usually only be one...)
		for ( auto& cameraEnt : cameras ){
			//TODO should be cliping this to only visible objects

			// enable required OpenGL state
			glEnable( GL_CULL_FACE );
			glCullFace( GL_BACK );

			// enable depth testing
			glEnable( GL_DEPTH_TEST );

			// set-up camera matrices
			auto* const camTransform = world.get<math::Transform>(cameraEnt);
			auto* const camComp = world.get<gfx::Camera>(cameraEnt);

			math::mat4 projectionMatrix = glm::perspective(camComp->fov, camComp->aspectRatio, camComp->nearPlane, camComp->farPlane);
			math::mat4 viewMatrix = glm::lookAt( camTransform->origin(), camComp->target, camTransform->up() );

			// TODO lighting needs to not be this...
			math::vec3 lightPos{20.0f, 20.0f, 15.0f};

			auto renderables = world.findMatching<StaticModel>();
			for ( const auto& entity : renderables ){
				auto* transform = world.get<math::Transform>(entity);
				StaticModel* model = world.get<StaticModel>(entity);

				// grouping by shader would mean we only need to send the model matrix...
				// TODO clean shader API
				auto shader = model->pipeline;
				if ( shader == nullptr ) {
					spdlog::warn("shader was null, aborting render");
					return;
				}

				shader->use();
				shader->setUniformMtx(shader->uniform("model"), transform->model());
				shader->setUniformMtx(shader->uniform("view"), viewMatrix);
				shader->setUniformMtx(shader->uniform("projection"), projectionMatrix);

				auto lightPosIdx = shader->uniform("lightPos");
				if ( lightPosIdx != -1 ) {
					shader->setUniformF(lightPosIdx, lightPos);
				}

				auto vao = model->vao;
				vao->bind();

				model->vertexData->bind();

				if ( model->restartIndex != NO_RESTART_IDX) {
					glEnable(GL_PRIMITIVE_RESTART);
					glPrimitiveRestartIndex(model->restartIndex);
				}
				auto elements = model->elements.get();
				vao->drawElements( *elements, model->drawType, model->elementCount);
				if ( model->restartIndex != NO_RESTART_IDX) {
					glDisable(GL_PRIMITIVE_RESTART);
				}
			}

		}
	}

}
